From: curt@eraserhead.jpl.nasa.gov
To:   All those who enjoy using CLIM

First off, I use a wide EMACs so excuse the wrapping, you may wish to reformat these forms at your
own leisure.  This code was derived from the source of with-output-to-output-record, with-new-output-record
and with-output-as-presentation for use by all those who use do not use CLIM's regular incremental
display methodology. In my particular case, I have my own scroll bars and scaling methodology so
I had generated presentations that were visible based upon my own scroll/pan positions and scales.
CLIM's incremental display was way too slow for 10k plus presentations and output-records, so I
sped up my display by not bothering with stuff that was not visible on the display.

Below is the code I quickly generated that allows you to recycle either presentations or output-records
within your own application frame.  If you are like me where you "cons up" many presentations and
output-records only to throw them away, this code will allow you to recycle them.  WARNING: I have only
run this in the CLIM 1.1 Franz Allegro World and these routines are probably not as efficient as they might
be, had one of the CLIM implementors provided for this mechanism.  I wish to thank Bill York at
York@lucid.com for supplying a mere CLIM user the macro source to attempt such a preverted thing!
I encourage the CLIM implementors to give some thought about recycling techniques on CLIM objects!

;;; -*- Mode: Common-Lisp; Package: Clim -*-

(in-package clim)

(defmacro Cycle-Output-Record
    ((stream old-record &optional record-type record &rest init-args) &body body)
  #+Genera (declare (zwei:indentation 0 3 1 1))    ;I kept has much code as possible within the original source
  (default-output-stream stream)
  ;; --- validate protocol here.
  `(with-output-recording-options (,stream :draw-p nil :record-p t)
     (letf-globally (((output-recording-stream-current-output-record-stack ,stream) nil)
		     ((output-recording-stream-output-record ,stream) nil)
		     ((output-recording-stream-text-output-record ,stream) nil))
		    (cond (,old-record (with-possibly-old-output-record (,stream ,old-record ,record-type ,record ,@init-args)
					 (with-stream-cursor-position-saved (,stream)
					   ,@body)))
			  (t (with-new-output-record (,stream ,record-type ,record ,@init-args)
			       (with-stream-cursor-position-saved (,stream)
				 ,@body)))))))


(defun with-old-output-record-internal (continuation stream old-record constructor
					&rest init-args &key parent &allow-other-keys)
  (declare (dynamic-extent init-args))
  (with-rem-keywords (init-args init-args '(:parent))
    (let* ((current-output-record (output-recording-stream-current-output-record-stack stream)))
      (multiple-value-bind (cursor-x cursor-y)
	  (stream-cursor-position* stream)
	(declare (fixnum cursor-x cursor-y))
	(multiple-value-bind (x y)
	    (multiple-value-bind (px py)
		(point-position*
		  (output-recording-stream-output-record-absolute-position stream))
	      (declare (fixnum px py))
	      (position-difference* cursor-x cursor-y px py))
	  (declare (fixnum x y))
	  (clear-output-record old-record)
	  (output-record-set-start-position* old-record x y)
	  (with-output-record-internal continuation stream old-record
				       cursor-x cursor-y)
	  ;; We set the parent after doing everything else so that calls
	  ;; to RECOMPUTE-CONTENTS-OK inside the dynamic extent of the
	  ;; continuation won't take forever.
	  (let ((parent (or parent
			    current-output-record
			    (output-recording-stream-output-record stream))))
	    (when parent (add-output-record-element parent old-record)))
	  old-record)))))


(defmacro with-possibly-old-output-record
    ((stream old-record &optional record-type record &rest init-args)
     &body body &environment env)
  #+Genera (declare (zwei:indentation 0 3 1 1))
  (unless record-type
    (setq record-type `'linear-output-record))
  (let ((constructor nil) (ignore-record nil))
    (when (and (constantp record-type #+(or Genera Minima) env)
	       (setq constructor
		 (gethash (eval record-type) *output-record-constructor-cache*))))
    (unless record
      (setq record '#:record
	    ignore-record t))
    `(flet ((with-new-output-record-body (,record)
	      ,@(when ignore-record `((declare (ignore ,record))))
	      ,@body))
       (with-old-output-record-internal		  
	   #'with-new-output-record-body
	 ,stream ,old-record ',constructor ,@init-args))))

(defmacro with-output-as-recycle-presentation ((&key object type stream modifier (old-presentation nil)
						     single-box (allow-sensitive-inferiors t)
						     (record-type `'standard-presentation)
						     parent)
					       &body body)
  #+Genera (declare (zwei:indentation 0 3 1 1))
  (default-output-stream stream)
  ;; Maybe with-new-output-record should turn record-p on?
  (let ((nobject '#:object)		;(once-only (object type) ...)
	(ntype '#:type))
    `(with-output-recording-options (,stream :record-p t)
       (let ((,nobject ,object)
	     (,ntype ,type))
	 (cond (,old-presentation
		(setf (presentation-object ,old-presentation) ,nobject
		      (presentation-type ,old-presentation) (if ,ntype (expand-presentation-type-abbreviation ,ntype)
							     (presentation-type-of ,nobject))
		      (presentation-single-box ,old-presentation) ,single-box)		
		(Cycle-Output-Record (,stream ,old-presentation ,record-type nil
						   :object ,nobject
						   :type (if ,ntype
							     (expand-presentation-type-abbreviation ,ntype)
							   (presentation-type-of ,nobject))
						   :single-box ,single-box
					      :allow-sensitive-inferiors ,allow-sensitive-inferiors
					      ,@(when modifier `(:modifier ,modifier))
					      ,@(when parent `(:parent ,parent)))
				     ,@body))
	       (t (with-new-output-record (,stream ,record-type nil
						   :object ,nobject
						   :type (if ,ntype
							     (expand-presentation-type-abbreviation ,ntype)
							   (presentation-type-of ,nobject))
						   :single-box ,single-box
						   :allow-sensitive-inferiors ,allow-sensitive-inferiors
						   ,@(when modifier `(:modifier ,modifier))
						   ,@(when parent `(:parent ,parent)))
		    ,@body)))))))

(defun my-present (object &optional (presentation-type (presentation-type-of object))
		&key (stream *standard-output*) (view (stream-default-view stream))
		     (modifier nil) (acceptably nil) (old-presentation nil)
		     (for-context-type presentation-type)
		     (single-box nil) (allow-sensitive-inferiors t)
		     (sensitive t) (record-type 'standard-presentation))

  ;; The arguments are allowed to be presentation type abbreviations
  (multiple-value-bind (expansion expanded)
      (expand-presentation-type-abbreviation presentation-type)
    (when expanded
      (when (eq for-context-type presentation-type)
	(setq for-context-type expansion))
      (setq presentation-type expansion)))
  (unless (eq for-context-type presentation-type)
    (multiple-value-bind (expansion expanded)
	(expand-presentation-type-abbreviation for-context-type)
      (when expanded
	(setq for-context-type expansion))))

  #+compulsive-type-checking
  (unless (presentation-typep object presentation-type)
    (cerror "Use the type ~*~*~S instead"
	    "The object ~S is not of type ~S"
	    object presentation-type (class-name (class-of object))))

;; Make a presentation if desired, and call the type's present method to
;; fill it in
  (if (and sensitive
	   ;; the right way to fix this is probably
	   ;; to make all the with-xxx macros turn into noops on non-window-streams, but
	   ;; this is easier and less expensive.
	   (extended-output-stream-p stream))
      (cond (old-presentation
	     (with-output-as-recycle-presentation (:stream stream :old-presentation old-presentation
							   :object object
							   :type presentation-type
							   :modifier modifier
							   :single-box single-box
							   :allow-sensitive-inferiors allow-sensitive-inferiors
							   :record-type record-type)
	       (call-presentation-generic-function present
						   object presentation-type stream view
						   :acceptably acceptably :for-context-type for-context-type)))
	    (t (with-output-as-presentation (:stream stream
						     :object object
						     :type presentation-type
						     :modifier modifier
						     :single-box single-box
						     :allow-sensitive-inferiors allow-sensitive-inferiors
						     :record-type record-type)
		 (call-presentation-generic-function present
						     object presentation-type stream view
						     :acceptably acceptably :for-context-type for-context-type))))
    (call-presentation-generic-function present
	object presentation-type stream view
	:acceptably acceptably :for-context-type for-context-type)))

;;--------------------------------------------------------------------
;Below is how I use these jewels to keep the gc at bay from my massive presentation and output-record
;generating program.
;ASSUMING the following:
;When I throw presentations or output-records away I keep track of them on a recycle list like
;For my presentations  (push presentation (get 'clim::standard-presentation 'old-standard-presentations))
;or for records (push output-record (get 'clim::linear-output-record 'old-output-records))
;You can now retrieve old presentations or records by simply popping them off lets say into OLD-ONE.
;So,

(with-output-as-recycle-presentation (:object whatever :stream stream :old-presentation OLD-ONE)
  (draw-a-presentation whatever stream)) ;to recycle an old presentation, instead of with-output-as-presentation

;or,

(cycle-output-record (stream OLD-ONE)
	;... body of code
		     )	;to recycle an old output-record, instead of using with-output-to-output-record,
			;or with-new-output-record!
